"""

identify_mislabeled_candidates.py

Identify images that may have been mislabeled.

A "mislabeled candidate" is defined as an image meeting both criteria:

* according to the ground-truth label, the model made an incorrect prediction

* the model's prediction confidence exceeds its confidence for the ground-truth
  label by at least <margin>

This script outputs for each dataset a text file containing the filenames of
mislabeled candidates, one per line. The text files are saved to:

    <logdir>/mislabeled_candidates_{split}_{dataset}.txt

To this list of files can then be passed to AzCopy to be downloaded:

""
azcopy cp "http://<url_of_container>?<sas_token>" "/save/files/here" \
       --list-of-files "/path/to/mislabeled_candidates_{split}_{dataset}.txt"
""

To save the filename as <dataset_name>/<blob_name> (instead of just <blob_name>
by default), pass the --include-dataset-in-filename flag. Then, the images can
be downloaded with:

""
  python data_management/megadb/download_images.py txt \
      "/path/to/mislabeled_candidates_{split}_{dataset}.txt" \
      /save/files/here \
      --threads 50
""

Assumes the following directory layout:
    <base_logdir>/
        label_index.json
        <logdir>/
            outputs_{split}.csv.gz

"""

#%% Imports

from __future__ import annotations

import argparse
from collections import defaultdict
from collections.abc import Iterable, Sequence
import json
import os

import pandas as pd
from tqdm import tqdm


#%% Example usage

"""
    python identify_mislabeled_candidates.py <base_logdir>/<logdir> \
        --margin 0.5 --splits val test
"""


#%% Main function

def main(logdir: str, splits: Iterable[str], margin: float,
         include_dataset_in_filename: bool) -> None:

    # load files
    logdir = os.path.normpath(logdir)  # removes any trailing slash
    base_logdir = os.path.dirname(logdir)
    idx_to_label_json_path = os.path.join(base_logdir, 'label_index.json')
    with open(idx_to_label_json_path, 'r') as f:
        idx_to_label = json.load(f)
    label_names = [idx_to_label[str(i)] for i in range(len(idx_to_label))]

    for split in splits:
        outputs_csv_path = os.path.join(logdir, f'outputs_{split}.csv.gz')
        candidates_df = get_candidates_df(outputs_csv_path, label_names, margin)

        # dataset => set of img_file
        candidate_image_files: defaultdict[str, set[str]] = defaultdict(set)

        for crop_path in tqdm(candidates_df['path']):
            # crop_path: <dataset>/<img_file>___cropXX_mdvY.Y.jpg
            #            [----<img_path>----]
            img_path = crop_path.split('___crop')[0]
            ds, img_file = img_path.split('/', maxsplit=1)
            if include_dataset_in_filename:
                candidate_image_files[ds].add(img_path)
            else:
                candidate_image_files[ds].add(img_file)

        for ds in sorted(candidate_image_files.keys()):
            img_files = candidate_image_files[ds]
            print(f'{ds} contains {len(img_files)} mislabeled candidates.')
            save_path = os.path.join(
                logdir, f'mislabeled_candidates_{split}_{ds}.txt')
            with open(save_path, 'w') as f:
                for img_file in sorted(img_files):
                    f.write(img_file + '\n')


#%% Support functions

def get_candidates_df(outputs_csv_path: str, label_names: Sequence[str],
                      margin: float) -> pd.DataFrame:
    """
    Returns a DataFrame containing crops only from mislabeled candidate
    images.
    """

    df = pd.read_csv(outputs_csv_path, float_precision='high')
    all_rows = range(len(df))
    df['pred'] = df[label_names].idxmax(axis=1)
    df['pred_conf'] = df.lookup(row_labels=all_rows, col_labels=df['pred'])
    df['label_conf'] = df.lookup(row_labels=all_rows, col_labels=df['label'])
    candidate_mask = df['pred_conf'] >= df['label_conf'] + margin
    candidates_df = df[candidate_mask].copy()
    return candidates_df


#%% Command-line driver

def _parse_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Identify mislabeled candidate images.')
    parser.add_argument(
        'logdir',
        help='folder inside <base_logdir> containing `outputs_<split>.csv.gz`')
    parser.add_argument(
        '--margin', type=float, default=0.5,
        help='confidence margin to count as a mislabeled candidate')
    parser.add_argument(
        '--splits', nargs='+', choices=['train', 'val', 'test'],
        help='which splits to identify mislabeled candidates on')
    parser.add_argument(
        '-d', '--include-dataset-in-filename', action='store_true',
        help='save filename as <dataset_name>/<blob_name>')
    return parser.parse_args()


if __name__ == '__main__':

    args = _parse_args()
    main(logdir=args.logdir, splits=args.splits, margin=args.margin,
         include_dataset_in_filename=args.include_dataset_in_filename)



"""

directory_listing.py

Script for creating Apache-style HTML directory listings for a local directory
and all its subdirectories.

Also includes a preview of a jpg file (the first in an alphabetical list),
if present.

"""

#%% Imports

import os
import sys
import argparse

from megadetector.utils.path_utils import is_image_file


#%% Directory enumeration functions

def _create_plain_index(root, dirs, files, dirname=None):
    """
    Creates the fairly plain HTML folder index including a preview of a single image file,
    if any is present.

    Args:
        root (str): path to the root directory, all paths in [dirs] and
            [files] are relative to this root folder
        dirs (list): list of strings, the directories in [root]
        files (list): list of strings, the files in [root]
        dirname (str, optional): name to print in the html,
            which may be different than [root]

    Returns:
        str: HTML source of the directory listing
    """

    if dirname is None:
        dirname = root or '/'
    dirname = dirname.replace('\\','/')

    html = "<!DOCTYPE html>\n"
    html += "<html lang='en'><head>"
    html += "<title>Index of {}</title>\n".format(dirname)
    html += "<meta charset='UTF-8'>\n"
    html += "<style>\n"
    html += "body { font-family: Segoe UI, Helvetica, Arial, sans-serif; }\n"
    html += "a { text-decoration:none; }\n"
    html += "</style>\n"
    html += "</head><body>\n"

    html += "<h1>Index of {}</h1>\n".format(dirname)

    # Insert preview image
    jpg_files = [f for f in files if is_image_file(f)]

    if len(jpg_files) > 0:

        # This is slow, so defaulting to just the first image:
        #
        # Use the largest image file as this is most likely to contain
        # useful content.
        #
        # jpg_file_sizes = [os.path.getsize(f) for f in jpg_files]
        # largest_file_index = max(range(len(jpg_files)), key=lambda x: jpg_file_sizes[x])

        html += "<a href='{0}'><img style='height:200px; float:right;' src='{0}' alt='Preview image'></a>\n".\
            format(jpg_files[0])
    else:
        html += "\n"
        # html += "<p style='width:15em; float:right; margin:0;'>[No preview available]</p>\n"

    if root:
        html += "<p><a href='../index.html'>To parent directory</a></p>\n"
    else:
        html += "\n"
        # html += "<p>This is the root directory.</p>\n"

    html += "<h2>Folders</h2>\n"
    if len(dirs) > 0:
        html += "<ul style='list-style-type: none; padding-left:1em;'>\n"
        for dir in sorted(dirs):
            html += "<li>&#128193; <a href='{0}/index.html'>{0}</a></li>\n".format(dir)
        html += "</ul>\n"
    else:
        html += "<p style='padding-left:1em;'>No folders</p>\n"

    html += "<h2>Files</h2>\n"
    if len(files) > 0:
        html += "<ul style='list-style-type: none; padding-left:1em;'>\n"
        for fname in sorted(files):
            if is_image_file(fname):
                html += "<li>&#128443; <a href='{0}'>{0}</a></li>\n".format(fname)
            else:
                html += "<li>&#128442; <a href='{0}'>{0}</a></li>\n".format(fname)
        html += "</ul>\n"
    else:
        html += "<p style = 'padding-left:1em;'>No files</p>\n"

    # Add some space at the bottom because the browser's status bar might hide stuff
    html += "<p style='margin:2em;'>&nbsp;</p>\n"
    html += "</body></html>\n"
    return html

# ...def _create_plain_index(...)


def create_html_index(dir,
                      overwrite=False,
                      template_fun=_create_plain_index,
                      basepath=None,
                      recursive=True):
    """
    Recursively traverses the local directory [dir] and generates a index
    file for each folder using [template_fun] to generate the HTML output.
    Excludes hidden files.

    Args:
        dir (str): directory to process
        overwrite (bool, optional): whether to over-write existing index file
        template_fun (func, optional): function taking three arguments (string,
            list of string, list of string) representing the current root, the list of folders,
            and the list of files.  Should return the HTML source of the index file.
        basepath (str, optional): if not None, the name used for each subfolder in [dir]
            in the output files will be relative to [basepath]
        recursive (bool, optional): recurse into subfolders
    """

    if template_fun is None:
        template_fun = _create_plain_index

    print('Traversing {}'.format(dir))

    # Make sure we remove the trailing /
    dir = os.path.normpath(dir)

    # Traverse directory and all sub directories, excluding hidden files
    for root, dirs, files in os.walk(dir):

        # Exclude files and folders that are hidden
        files = [f for f in files if not f[0] == '.']
        dirs[:] = [d for d in dirs if not d[0] == '.']

        # Output is written to file *root*/index.html
        output_file = os.path.join(root, "index.html")

        if (not overwrite) and os.path.isfile(output_file):
            print('Skipping {}, file exists'.format(output_file))
            continue

        print("Generating {}".format(output_file))

        # Generate HTML with template function
        dirname = None
        if basepath is not None:
            dirname = os.path.relpath(root,basepath)
        html = template_fun(root[len(dir):], dirs, files, dirname)

        # Write to file
        with open(output_file, 'wt') as fi:
            fi.write(html)

        if not recursive:
            break

# ...def create_html_index(...)


#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser()

    parser.add_argument("directory", type=str,
                        help='Path to directory which should be traversed.')
    parser.add_argument("--basepath", type=str,
                        help='Folder names will be printed relative to basepath, if specified',
                        default=None)
    parser.add_argument("--overwrite", action='store_true', default=False,
                        help='If set, the script will overwrite existing index.html files.')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    assert os.path.isdir(args.directory), "{} is not a valid directory".format(args.directory)

    create_html_index(args.directory,
                      overwrite=args.overwrite,
                      basepath=args.basepath)

if __name__ == '__main__':
    main()



"""

remap_coco_categories.py

Given a COCO-formatted dataset, remap the categories to a new mapping.  A common use
case is to take a fine-grained dataset (e.g. with species-level categories) and
map them to coarse categories (typically MD categories).

"""

#%% Imports and constants

import os
import json
import argparse

from copy import deepcopy
from megadetector.utils.ct_utils import invert_dictionary


#%% Main function

def remap_coco_categories(input_data,
                          output_category_name_to_id,
                          input_category_name_to_output_category_name,
                          output_file=None,
                          allow_unused_categories=False):
    """
    Given a COCO-formatted dataset, remap the categories to a new categories mapping, optionally
    writing the results to a new file.

    Args:
        input_data (str or dict): a COCO-formatted dict or a filename.  If it's a dict, it will
            be copied, not modified in place.
        output_category_name_to_id (dict): a dict mapping strings to ints.  Categories not in
            this dict will be ignored or will result in errors, depending on allow_unused_categories.
        input_category_name_to_output_category_name (dict): a dict mapping strings to strings.
            Annotations using categories not in this dict will be omitted or will result in
            errors, depending on allow_unused_categories.
        output_file (str, optional): output file to which we should write remapped COCO data
        allow_unused_categories (bool, optional): should we ignore categories not present in the
            input/output mappings?  If this is False and we encounter an unmapped category, we'll
            error.

    Returns:
        dict: COCO-formatted dict
    """

    if isinstance(input_data,str):
        assert os.path.isfile(input_data), "Can't find file {}".format(input_data)
        with open(input_data,'r') as f:
            input_data = json.load(f)
        assert isinstance(input_data,dict), 'Illegal COCO input data'
    else:
        assert isinstance(input_data,dict), 'Illegal COCO input data'
        input_data = deepcopy(input_data)

    # It's safe to modify in-place now
    output_data = input_data

    # Read input name --> ID mapping
    input_category_name_to_input_category_id = {}
    for c in input_data['categories']:
        input_category_name_to_input_category_id[c['name']] = c['id']
    input_category_id_to_input_category_name = \
        invert_dictionary(input_category_name_to_input_category_id)

    # Map input IDs --> output IDs
    input_category_id_to_output_category_id = {}
    input_category_names = list(input_category_name_to_output_category_name.keys())

    # input_name = input_category_names[0]
    for input_name in input_category_names:

        output_name = input_category_name_to_output_category_name[input_name]
        assert output_name in output_category_name_to_id, \
            'No output ID for {} --> {}'.format(input_name,output_name)
        input_id = input_category_name_to_input_category_id[input_name]
        output_id = output_category_name_to_id[output_name]
        input_category_id_to_output_category_id[input_id] = output_id

    # ...for each category we want to keep

    printed_unused_category_warnings = set()

    valid_annotations = []

    # Map annotations
    for ann in output_data['annotations']:

        input_category_id = ann['category_id']
        if input_category_id not in input_category_id_to_output_category_id:
            if allow_unused_categories:
                if input_category_id not in printed_unused_category_warnings:
                    printed_unused_category_warnings.add(input_category_id)
                    input_category_name = \
                        input_category_id_to_input_category_name[input_category_id]
                    s = 'Skipping unmapped category ID {} ({})'.format(
                        input_category_id,input_category_name)
                    print(s)
                continue
            else:
                s = 'Unmapped category ID {}'.format(input_category_id)
                raise ValueError(s)
        output_category_id = input_category_id_to_output_category_id[input_category_id]
        ann['category_id'] = output_category_id
        valid_annotations.append(ann)

    # ...for each annotation

    # The only reason annotations should get excluded is the case where we allow
    # unused categories
    if not allow_unused_categories:
        assert len(valid_annotations) == len(output_data['annotations'])

    output_data['annotations'] = valid_annotations

    # Update the category list
    output_categories = []
    for output_name in output_category_name_to_id:
        category = {'name':output_name,'id':output_category_name_to_id[output_name]}
        output_categories.append(category)
    output_data['categories'] = output_categories

    if output_file is not None:
        with open(output_file,'w') as f:
            json.dump(output_data,f,indent=1)

    return output_data

# ...def remap_coco_categories(...)


#%% Command-line driver

def main():
    """
    Command-line interface to remap COCO categories.
    """

    parser = argparse.ArgumentParser(
        description='Remap categories in a COCO-formatted dataset'
    )
    parser.add_argument(
        'input_coco_file',
        type=str,
        help='Path to the input COCO .json file'
    )
    parser.add_argument(
        'output_category_map_file',
        type=str,
        help="Path to a .json file mapping output category names to integer IDs (e.g., {'cat':0, 'dog':1})"
    )
    parser.add_argument(
        'input_to_output_category_map_file',
        type=str,
        help="Path to a .json file mapping input category names to output category names" + \
            " (e.g., {'old_cat_name':'cat', 'old_dog_name':'dog'})"
    )
    parser.add_argument(
        'output_coco_file',
        type=str,
        help='Path to save the remapped COCO .json file'
    )
    parser.add_argument(
        '--allow_unused_categories',
        action='store_true',
        help='Allow unmapped categories (by default, errors on unmapped categories)'
    )

    args = parser.parse_args()

    # Load category mappings
    with open(args.output_category_map_file, 'r') as f:
        output_category_name_to_id = json.load(f)

    with open(args.input_to_output_category_map_file, 'r') as f:
        input_category_name_to_output_category_name = json.load(f)

    # Load COCO data
    with open(args.input_coco_file, 'r') as f:
        input_data = json.load(f)

    remap_coco_categories(
        input_data=input_data,
        output_category_name_to_id=output_category_name_to_id,
        input_category_name_to_output_category_name=input_category_name_to_output_category_name,
        output_file=args.output_coco_file,
        allow_unused_categories=args.allow_unused_categories
    )

    print(f'Successfully remapped categories and saved to {args.output_coco_file}')

if __name__ == '__main__':
    main()
